@@ -1,5 +1,5 @@
 /*
- *	$Id: lspci.c,v 1.6 1998/02/08 10:58:54 mj Exp $
+ *	$Id: lspci.c,v 1.12 1998/06/08 07:54:37 mj Exp $
  *
  *	Linux PCI Utilities -- List All PCI Devices
  *
@@ -13,7 +13,6 @@
 #include <stdlib.h>
 #include <fcntl.h>
 #include <unistd.h>
-#include <linux/pci.h>
 
 #include "pciutils.h"
 
@@ -22,25 +21,24 @@
 static int verbose;			/* Show detailed information */
 static int buscentric_view;		/* Show bus addresses/IRQ's instead of CPU-visible ones */
 static int show_hex;			/* Show contents of config space as hexadecimal numbers */
-static int bus_filter = -1;		/* Bus, slot, function, vendor and device ID filtering */
-static int slot_filter = -1;
-static int func_filter = -1;
-static int vend_filter = -1;
-static int dev_filter = -1;
+static struct pci_filter filter;	/* Device filter */
 static int show_tree;			/* Show bus tree */
+static int machine_readable;		/* Generate machine-readable output */
 static char *pci_dir = PROC_BUS_PCI;
 
-static char options[] = "nvbxB:S:F:V:D:ti:p:";
+static char options[] = "nvbxs:d:ti:p:m";
 
 static char help_msg[] = "\
 Usage: lspci [<switches>]\n\
 \n\
--v\tBe verbose\n\
--n\tShow numeric ID's\n\
--b\tBus-centric view (PCI addresses and IRQ's instead of those seen by the CPU)\n\
--x\tShow hex-dump of config space (-xx shows full 256 bytes)\n\
--B <bus>, -S <slot>, -F <func>, -V <vendor>, -D <device>  Show only selected devices\n\
--t\tShow bus tree\n\
+-v\t\tBe verbose\n\
+-n\t\tShow numeric ID's\n\
+-b\t\tBus-centric view (PCI addresses and IRQ's instead of those seen by the CPU)\n\
+-x\t\tShow hex-dump of config space (-xx shows full 256 bytes)\n\
+-s [[<bus>]:][<slot>][.[<func>]]\tShow only devices in selected slots\n\
+-d [<vendor>]:[<device>]\tShow only selected devices\n\
+-t\t\tShow bus tree\n\
+-m\t\tProduce machine-readable output\n\
 -i <file>\tUse specified ID database instead of " ETC_PCI_IDS "\n\
 -p <dir>\tUse specified bus directory instead of " PROC_BUS_PCI "\n\
 ";
@@ -60,7 +58,7 @@ struct device {
   byte bus, devfn;
   word vendid, devid;
   unsigned int kernel_irq;
-  unsigned long kernel_base_addr[6];
+  unsigned long kernel_base_addr[6], kernel_rom_base_addr;
   byte config[256];
 };
 
@@ -80,18 +78,6 @@ xmalloc(unsigned int howmuch)
   return p;
 }
 
-/* Filtering */
-
-static inline int
-filter_out(struct device *d)
-{
-  return (bus_filter >= 0 && d->bus != bus_filter ||
-	  slot_filter >= 0 && PCI_SLOT(d->devfn) != slot_filter ||
-	  func_filter >= 0 && PCI_FUNC(d->devfn) != func_filter ||
-	  vend_filter >= 0 && d->vendid != vend_filter ||
-	  dev_filter >= 0 && d->devid != dev_filter);
-}
-
 /* Interface for /proc/bus/pci */
 
 static void
@@ -112,7 +98,8 @@ scan_dev_list(void)
       struct device *d = xmalloc(sizeof(struct device));
       unsigned int dfn, vend;
 
-      sscanf(line, "%x %x %x %lx %lx %lx %lx %lx %lx",
+      bzero(d, sizeof(*d));
+      sscanf(line, "%x %x %x %lx %lx %lx %lx %lx %lx %lx",
 	     &dfn,
 	     &vend,
 	     &d->kernel_irq,
@@ -121,12 +108,13 @@ scan_dev_list(void)
 	     &d->kernel_base_addr[2],
 	     &d->kernel_base_addr[3],
 	     &d->kernel_base_addr[4],
-	     &d->kernel_base_addr[5]);
+	     &d->kernel_base_addr[5],
+	     &d->kernel_rom_base_addr);
       d->bus = dfn >> 8U;
       d->devfn = dfn & 0xff;
       d->vendid = vend >> 16U;
       d->devid = vend & 0xffff;
-      if (!filter_out(d))
+      if (filter_match(&filter, d->bus, d->devfn, d->vendid, d->devid))
 	{
 	  *last_dev = d;
 	  last_dev = &d->next;
@@ -327,16 +315,13 @@ show_bases(struct device *d, int cnt)
 static void
 show_htype0(struct device *d)
 {
-  u32 rom = get_conf_long(d, PCI_ROM_ADDRESS);
+  unsigned long rom = buscentric_view ? get_conf_long(d, PCI_ROM_ADDRESS) : d->kernel_rom_base_addr;
 
   show_bases(d, 6);
 
   if (rom & 1)
-    {
-      word cmd = get_conf_word(d, PCI_COMMAND);
-      printf("\tExpansion ROM at %08x%s\n", rom & ~0xfff,
-	     (cmd & PCI_COMMAND_MEMORY) ? "" : " [disabled]");
-    }
+    printf("\tExpansion ROM at %08lx%s\n", rom & PCI_ROM_ADDRESS_MASK,
+	   (rom & PCI_ROM_ADDRESS_ENABLE) ? "" : " [disabled]");
 }
 
 static void
@@ -351,7 +336,7 @@ show_htype1(struct device *d)
   u32 pref_base = get_conf_word(d, PCI_PREF_MEMORY_BASE);
   u32 pref_limit = get_conf_word(d, PCI_PREF_MEMORY_LIMIT);
   u32 pref_type = pref_base & PCI_PREF_RANGE_TYPE_MASK;
-  u32 rom = get_conf_long(d, PCI_ROM_ADDRESS1);
+  unsigned long rom = buscentric_view ? get_conf_long(d, PCI_ROM_ADDRESS) : d->kernel_rom_base_addr;
   word brc = get_conf_word(d, PCI_BRIDGE_CONTROL);
 
   show_bases(d, 2);
@@ -408,11 +393,8 @@ show_htype1(struct device *d)
     printf("\tSecondary status: SERR\n");
 
   if (rom & 1)
-    {
-      word cmd = get_conf_word(d, PCI_COMMAND);
-      printf("\tExpansion ROM at %08x%s\n", rom & ~0xfff,
-	     (cmd & PCI_COMMAND_MEMORY) ? "" : " [disabled]");
-    }
+    printf("\tExpansion ROM at %08lx%s\n", rom & PCI_ROM_ADDRESS_MASK,
+	   (rom & PCI_ROM_ADDRESS_ENABLE) ? "" : " [disabled]");
 
   if (verbose > 1)
     printf("\tBridgeCtl: Parity%c SERR%c NoISA%c VGA%c MAbort%c >Reset%c FastB2B%c\n",
@@ -594,6 +576,37 @@ show_hex_dump(struct device *d)
     }
 }
 
+static void
+show_machine(struct device *d)
+{
+  int c;
+
+  if (verbose)
+    {
+      printf("Device:\t%02x:%02x.%x\n", d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+      printf("Class:\t%s\n", lookup_class(get_conf_word(d, PCI_CLASS_DEVICE)));
+      printf("Vendor:\t%s\n", lookup_vendor(d->vendid));
+      printf("Device:\t%s\n", lookup_device(d->vendid, d->devid));
+      if (c = get_conf_byte(d, PCI_REVISION_ID))
+	printf("Rev:\t%02x\n", c);
+      if (c = get_conf_byte(d, PCI_CLASS_PROG))
+	printf("ProgIf:\t%02x\n", c);
+    }
+  else
+    {
+      printf("%02x:%02x.%x ", d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+      printf("\"%s\" \"%s\" \"%s\"",
+	     lookup_class(get_conf_word(d, PCI_CLASS_DEVICE)),
+	     lookup_vendor(d->vendid),
+	     lookup_device(d->vendid, d->devid));
+      if (c = get_conf_byte(d, PCI_REVISION_ID))
+	printf(" -r%02x", c);
+      if (c = get_conf_byte(d, PCI_CLASS_PROG))
+	printf(" -p%02x", c);
+      putchar('\n');
+    }
+}
+
 static void
 show(void)
 {
@@ -601,7 +614,9 @@ show(void)
 
   for(d=first_dev; d; d=d->next)
     {
-      if (verbose)
+      if (machine_readable)
+	show_machine(d);
+      else if (verbose)
 	show_verbose(d);
       else
 	show_terse(d);
@@ -681,11 +696,11 @@ static void
 grow_tree(void)
 {
   struct device *d, *d2;
-  struct bridge *first_br, *b;
+  struct bridge **last_br, *b;
 
   /* Build list of bridges */
 
-  first_br = &host_bridge;
+  last_br = &host_bridge.chain;
   for(d=first_dev; d; d=d->next)
     {
       word class = get_conf_word(d, PCI_CLASS_DEVICE);
@@ -695,21 +710,22 @@ grow_tree(void)
 	  b->primary = get_conf_byte(d, PCI_PRIMARY_BUS);
 	  b->secondary = get_conf_byte(d, PCI_SECONDARY_BUS);
 	  b->subordinate = get_conf_byte(d, PCI_SUBORDINATE_BUS);
-	  b->chain = first_br;
-	  first_br = b;
+	  *last_br = b;
+	  last_br = &b->chain;
 	  b->next = b->child = NULL;
 	  b->first_bus = NULL;
 	  b->br_dev = d;
 	}
     }
+  *last_br = NULL;
 
   /* Create a bridge tree */
 
-  for(b=first_br; b; b=b->chain)
+  for(b=&host_bridge; b; b=b->chain)
     {
       struct bridge *c, *best;
       best = NULL;
-      for(c=first_br; c; c=c->chain)
+      for(c=&host_bridge; c; c=c->chain)
 	if (c != b && b->primary >= c->secondary && b->primary <= c->subordinate &&
 	    (!best || best->subordinate - best->primary > c->subordinate - c->primary))
 	  best = c;
@@ -722,7 +738,7 @@ grow_tree(void)
 
   /* Insert secondary bus for each bridge */
 
-  for(b=first_br; b; b=b->chain)
+  for(b=&host_bridge; b; b=b->chain)
     if (!find_bus(b, b->secondary))
       new_bus(b, b->secondary);
 
@@ -743,7 +759,7 @@ print_it(byte *line, byte *p)
   *p = 0;
   fputs(line, stdout);
   for(p=line; *p; p++)
-    if (*p == '+')
+    if (*p == '+' || *p == '|')
       *p = '|';
     else
       *p = ' ';
@@ -760,7 +776,10 @@ show_tree_dev(struct device *d, byte *line, byte *p)
   for(b=&host_bridge; b; b=b->chain)
     if (b->br_dev == d)
       {
-	p += sprintf(p, "-[%02x-%02x]-", b->secondary, b->subordinate);
+	if (b->secondary == b->subordinate)
+	  p += sprintf(p, "-[%02x]-", b->secondary);
+	else
+	  p += sprintf(p, "-[%02x-%02x]-", b->secondary, b->subordinate);
         show_tree_bridge(b, line, p);
         return;
       }
@@ -837,7 +856,14 @@ int
 main(int argc, char **argv)
 {
   int i;
+  char *msg;
 
+  if (argc == 2 && !strcmp(argv[1], "--version"))
+    {
+      puts("lspci version " PCIUTILS_VERSION);
+      return 0;
+    }
+  filter_init(&filter);
   while ((i = getopt(argc, argv, options)) != -1)
     switch (i)
       {
@@ -850,20 +876,19 @@ main(int argc, char **argv)
       case 'b':
 	buscentric_view = 1;
 	break;
-      case 'B':
-	bus_filter = strtol(optarg, NULL, 16);
+      case 's':
+	if (msg = filter_parse_slot(&filter, optarg))
+	  {
+	    fprintf(stderr, "lspci: -f: %s\n", msg);
+	    return 1;
+	  }
 	break;
-      case 'S':
-	slot_filter = strtol(optarg, NULL, 16);
-	break;
-      case 'F':
-	func_filter = strtol(optarg, NULL, 16);
-	break;
-      case 'V':
-	vend_filter = strtol(optarg, NULL, 16);
-	break;
-      case 'D':
-	dev_filter = strtol(optarg, NULL, 16);
+      case 'd':
+	if (msg = filter_parse_id(&filter, optarg))
+	  {
+	    fprintf(stderr, "lspci: -d: %s\n", msg);
+	    return 1;
+	  }
 	break;
       case 'x':
 	show_hex++;
@@ -877,6 +902,9 @@ main(int argc, char **argv)
       case 'p':
 	pci_dir = optarg;
 	break;
+      case 'm':
+	machine_readable++;
+	break;
       default:
       bad:
 	fprintf(stderr, help_msg);
